{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n\nFormat STARmap data\n===================\n\nThe raw STARmap data received by SpaceTx is organized as z-stack TIFF files that contain all\nz-sections for each (round, channel) pair. This TileFetcher breaks these up into individual 2-d\nTIFF files, reformats the CSV codebook, and adds the necessary SpaceTx metadata.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import functools\nimport os\nfrom typing import Mapping, Tuple, Union\n\nimport click\nimport numpy as np\nfrom skimage.io import imread\nfrom slicedimage import ImageFormat\n\nfrom starfish import Codebook\nfrom starfish.experiment.builder import FetchedTile, TileFetcher, write_experiment_json\nfrom starfish.types import Axes, Coordinates, Features\n\n\n# We use this to cache images. This allows us to write a separate function to extract the data's\n# shape from the file, instead of hard-coding the shape. It also enables us to call the function\n# once per z-slice, but only read the 3d volume one time.\n@functools.lru_cache(maxsize=1)\ndef cached_read_fn(file_path) -> np.ndarray:\n    return imread(file_path)\n\n\nclass StarMapTile(FetchedTile):\n\n    def __init__(\n            self,\n            file_path: str,\n            z: int\n    ) -> None:\n        \"\"\"Parser for an osmFISH tile.\n\n        Parameters\n        ----------\n        file_path : str\n            location of the StarMap 3d-TIFF\n        coordinates :\n            the coordinates for the selected StarMap tile, extracted from the metadata\n        z : int\n            the z-layer to extract from the StarMap tile\n        \"\"\"\n        self.file_path = file_path\n        self.z = z\n\n        # dummy coordinates\n        self._coordinates = {\n            Coordinates.X: (0.0, 0.0001),\n            Coordinates.Y: (0.0, 0.0001),\n            Coordinates.Z: (0.0, 0.0001),\n        }\n\n    @property\n    def shape(self) -> Mapping[Axes, int]:\n        return {Axes.Y: 3738, Axes.X: 17247}  # hard coded for these datasets.\n\n    @property\n    def coordinates(self):\n        return self._coordinates  # noqa\n\n    def tile_data(self) -> np.ndarray:\n        return cached_read_fn(self.file_path)[self.z]  # slice out the correct z-plane\n\n\nclass StarMapTileFetcher(TileFetcher):\n\n    def __init__(self, input_dir: str) -> None:\n        \"\"\"Implement a TileFetcher for a StarMap experiment.\n\n        This TileFetcher constructs spaceTx format for a stitched tissue slice, where\n        `input_dir` is a directory containing .tif files with the following structure:\n\n        \u251c\u2500\u2500 genes.csv\n        \u2514\u2500\u2500 reg3d\n            \u251c\u2500\u2500 reg_round01.tif\n            \u251c\u2500\u2500 reg_round01_ch01.tif\n            \u251c\u2500\u2500 reg_round01_ch02.tif\n            \u251c\u2500\u2500 reg_round01_ch03.tif\n            \u251c\u2500\u2500 reg_round01_ch04.tif\n            \u251c\u2500\u2500 reg_round02.tif\n            \u251c\u2500\u2500 reg_round02_ch01.tif\n            \u251c\u2500\u2500 reg_round02_ch02.tif\n            \u251c\u2500\u2500 reg_round02_ch03.tif\n            \u251c\u2500\u2500 reg_round02_ch04.tif\n            \u251c\u2500\u2500 reg_round03.tif\n            \u251c\u2500\u2500 reg_round03_ch01.tif\n            \u251c\u2500\u2500 reg_round03_ch02.tif\n            \u251c\u2500\u2500 reg_round03_ch03.tif\n            \u251c\u2500\u2500 reg_round03_ch04.tif\n            \u251c\u2500\u2500 reg_round04.tif\n            \u251c\u2500\u2500 reg_round04_ch01.tif\n            \u251c\u2500\u2500 reg_round04_ch02.tif\n            \u251c\u2500\u2500 reg_round04_ch03.tif\n            \u251c\u2500\u2500 reg_round04_ch04.tif\n            \u251c\u2500\u2500 reg_round05.tif\n            \u251c\u2500\u2500 reg_round05_ch01.tif\n            \u251c\u2500\u2500 reg_round05_ch02.tif\n            \u251c\u2500\u2500 reg_round05_ch03.tif\n            \u251c\u2500\u2500 reg_round05_ch04.tif\n            \u251c\u2500\u2500 reg_round06.tif\n            \u251c\u2500\u2500 reg_round06_ch01.tif\n            \u251c\u2500\u2500 reg_round06_ch02.tif\n            \u251c\u2500\u2500 reg_round06_ch03.tif\n            \u2514\u2500\u2500 reg_round06_ch04.tif\n\n        Notes\n        -----\n        - The spatial organization of the fields of view are not known to the starfish developers,\n          so they are filled by dummy coordinates\n        - This TileFetcher is specifically tailored to the gene panel used for a specific\n          experiment. Generalization of this TileFetcher will require reimplementation of the\n          `channel_map` method.\n        \"\"\"\n\n        self.input_dir = input_dir\n        self.num_z = 28  # hard coded for this dataset\n\n    def get_tile(\n            self, fov_id: int, round_label: int, ch_label: int, zplane_label: int) -> FetchedTile:\n        basename = f\"reg_round0{round_label+ 1}_ch0{ch_label + 1}.tif\"  # translate to 3d\n        file_path = os.path.join(self.input_dir, \"reg3d\", basename)\n        return StarMapTile(file_path, zplane_label)\n\n    def generate_codebook(self, output_dir: str) -> None:\n        \"\"\"Generate and save a codebook from the provided mapping of genes to DNA sequences.\n\n        StarMAP codebooks are encoded with the 2-base encoding used for solid sequencing. In this\n        scheme, multiple pairs of bases map into the same fluorescence channels. This function\n        exposes this mapping.\n\n        Parameters\n        ----------\n        output_dir : str\n            directory in which to save the generated codebook. Codebook is saved as \"codebook.json\"\n\n        \"\"\"\n        dinucleotides_to_channels = {\n            \"AT\": 4,\n            \"CT\": 3,\n            \"GT\": 2,\n            \"TT\": 1,\n            \"AG\": 3,\n            \"CG\": 4,\n            \"GG\": 1,\n            \"TG\": 2,\n            \"AC\": 2,\n            \"CC\": 1,\n            \"GC\": 4,\n            \"TC\": 3,\n            \"AA\": 1,\n            \"CA\": 2,\n            \"GA\": 3,\n            \"TA\": 4,\n        }\n\n        with open(os.path.join(self.input_dir, \"genes.csv\"), \"r\") as f:\n            codes = [l.strip().split(\",\") for l in f.readlines()]  # List[(gene, dna_barcode), ...]\n\n        def iter_dinucleotides(sequence):\n            i = 0\n            while i + 1 < len(sequence):\n                yield sequence[i:i + 2]\n                i += 1\n\n        # construct codebook target mappings\n        code_array = []\n        for gene, dna_barcode in codes:\n            dna_barcode = dna_barcode[::-1]  # reverse barcode\n            spacetx_barcode = [\n                {\n                    Axes.ROUND.value: r,\n                    Axes.CH.value: dinucleotides_to_channels[dinucleotide],\n                    Features.CODE_VALUE: 1\n                } for r, dinucleotide in enumerate(iter_dinucleotides(dna_barcode))\n            ]\n            code_array.append({\n                Features.CODEWORD: spacetx_barcode,\n                Features.TARGET: gene\n            })\n\n        codebook = Codebook.from_code_array(code_array)\n        codebook.to_json(os.path.join(output_dir, \"codebook.json\"))\n\n\nclass StarMapDapiTileFetcher(TileFetcher):\n\n    def __init__(self, input_dir: str) -> None:\n        \"\"\"Implement a TileFetcher for dapi auxiliary images of a StarMap experiment.\n\n\n        \u2514\u2500\u2500 nissl\n         \u00a0\u00a0 \u251c\u2500\u2500 dapi.tif\n         \u00a0\u00a0 \u251c\u2500\u2500 dapi_maxproj.tif\n\n        \"\"\"\n        self.input_dir = input_dir\n\n    def get_tile(\n            self, fov_id: int, round_label: int, ch_label: int, zplane_label: int) -> FetchedTile:\n        basename = f\"reg_round0{round_label + 1}_ch0{ch_label + 1}.tif\"\n        file_path = os.path.join(self.input_dir, \"reg3d\", basename)\n        return StarMapTile(file_path, zplane_label)\n\n\nclass StarMapNisslTileFetcher(TileFetcher):\n\n    def __init__(self, input_dir: str, channels: Tuple[int, int]) -> None:\n        \"\"\"Implement a TileFetcher for nissl auxiliary images of a StarMap experiment.\n\n        Images are expected to be found with the following names. Since the nissl images\n        correspond to projections across pairs of channels, a separate TileFetcher will be\n        initialized for each projection.\n\n        input_dir\n        \u2514\u2500\u2500 nissl\n         \u00a0\u00a0 \u251c\u2500\u2500 2base_BF2_nissl_decon_ch00.tif\n         \u00a0\u00a0 \u251c\u2500\u2500 2base_BF2_nissl_decon_ch01.tif\n         \u00a0\u00a0 \u251c\u2500\u2500 2base_BF2_nissl_decon_ch02.tif\n         \u00a0\u00a0 \u251c\u2500\u2500 nissl.tif\n         \u00a0\u00a0 \u2514\u2500\u2500 nissl_maxproj.tif\n\n        Parameters\n        ----------\n        input_dir : str\n            directory containing input data\n        channels : Tuple[int, int]\n            the pair of channels corresponding to the TIFF file to be read.\n\n        \"\"\"\n        self.input_dir = input_dir\n\n    def get_tile(\n            self, fov_id: int, round_label: int, ch_label: int, zplane_label: int) -> FetchedTile:\n        basename = f\"reg_round0{round_label + 1}_ch0{ch_label + 1}.tif\"\n        file_path = os.path.join(self.input_dir, \"reg3d\", basename)\n        return StarMapTile(file_path, zplane_label)\n\n\n@click.command()\n@click.option(\"--input-dir\", type=str, required=True, help=\"input directory containing images\")\n@click.option(\"--output-dir\", type=str, required=True, help=\"output directory for formatted data\")\ndef cli(input_dir, output_dir) -> None:\n    \"\"\"CLI entrypoint for spaceTx format construction for osmFISH data\n\n    Raw data (input for this tool) for this experiment can be found at:\n    s3://spacetx.starfish.data.public/browse/raw/20181031/starmap/\n\n    Processed data (output of this tool) can be found at:\n    s3://spacetx.starfish.data.public/browse/formatted/20190111/starmap/\n\n    Parameters\n    ----------\n    input_dir : str\n        directory containing input data. See TileFetcher classes for expected directory structures.\n        Example data can be found at\n        s3://spacetx.starfish.data/browse/raw/20183110/starmap/160gene_20171110_BF2_dark/\n    output_dir : str\n        directory that 2-d images and SpaceTx metadata will be written to.\n    \"\"\"\n    abs_output_dir = os.path.expanduser(output_dir)\n    abs_input_dir = os.path.expanduser(input_dir)\n    os.makedirs(abs_output_dir, exist_ok=True)\n\n    primary_tile_fetcher = StarMapTileFetcher(abs_input_dir)\n    dapi_tile_fetcher = StarMapDapiTileFetcher(abs_input_dir)\n    nissl_00_tile_fetcher = StarMapNisslTileFetcher(abs_input_dir, channels=(0, 0))\n    nissl_01_tile_fetcher = StarMapNisslTileFetcher(abs_input_dir, channels=(0, 1))\n    nissl_02_tile_fetcher = StarMapNisslTileFetcher(abs_input_dir, channels=(0, 2))\n\n    # This is hardcoded for this example data set\n    primary_image_dimensions: Mapping[Union[str, Axes], int] = {\n        Axes.ROUND: 6,\n        Axes.CH: 4,\n        Axes.ZPLANE: 28,\n    }\n\n    aux_images_dimensions: Mapping[str, Mapping[Union[str, Axes], int]] = {\n        \"nuclei\": {\n            Axes.ROUND: 1,\n            Axes.CH: 1,\n            Axes.ZPLANE: 1,\n        },\n        \"nissl_channels_0_0\": {\n            Axes.ROUND: 1,\n            Axes.CH: 1,\n            Axes.ZPLANE: 1,\n        },\n        \"nissl_channels_0_1\": {\n            Axes.ROUND: 1,\n            Axes.CH: 1,\n            Axes.ZPLANE: 1,\n        },\n        \"nissl_channels_0_2\": {\n            Axes.ROUND: 1,\n            Axes.CH: 1,\n            Axes.ZPLANE: 1,\n        },\n    }\n\n    write_experiment_json(\n        path=output_dir,\n        fov_count=1,\n        tile_format=ImageFormat.TIFF,\n        primary_image_dimensions=primary_image_dimensions,\n        aux_name_to_dimensions=aux_images_dimensions,\n        primary_tile_fetcher=primary_tile_fetcher,\n        aux_tile_fetcher={\n            \"nuclei\": dapi_tile_fetcher,\n            \"nissl_channels_0_0\": nissl_00_tile_fetcher,\n            \"nissl_channels_0_1\": nissl_01_tile_fetcher,\n            \"nissl_hcannels_0_2\": nissl_02_tile_fetcher\n        },\n        dimension_order=(Axes.ROUND, Axes.CH, Axes.ZPLANE)\n    )\n\n    primary_tile_fetcher.generate_codebook(abs_output_dir)\n\n\nif __name__ == \"__main__\":\n    cli()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.6.8"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}